
import numpy as np
from typing import Optional


def fixed_value_threshold(ground_truth: np.array, scores: np.array, threshold: Optional[float] = None) -> np.array:
    # If no threshold is given, use the ground truth number of anomalies to compute a threshold (which boils down to top_n_threshold)
    if threshold is None:
        return top_n_threshold(ground_truth, scores)

    if threshold <= 0.0:
        return np.ones_like(scores, dtype=np.int16)
    elif threshold >= 1.0:
        return np.zeros_like(scores, dtype=np.int16)
    else:
        return np.array(scores >= threshold, dtype=np.int16)


def contamination_threshold(ground_truth: np.array, scores: np.array, contamination: Optional[float] = None) -> np.array:
    # If no contamination is given, compute the contamination based on the ground truth
    if contamination is None:
        contamination = np.sum(ground_truth) / len(ground_truth)

    if contamination <= 0.0:
        return np.zeros_like(scores, dtype=np.int16)
    elif contamination >= 1.0:
        return np.ones_like(scores, dtype=np.int16)
    else:
        return np.array(scores >= np.quantile(scores, 1 - contamination), dtype=np.int16)


def top_n_threshold(ground_truth: np.array, scores: np.array, top_n: Optional[int] = None) -> np.array:
    # If no n is given, use the ground truth number of anomalies to compute a threshold
    if top_n is None:
        top_n = np.sum(ground_truth)
    return contamination_threshold(ground_truth, scores, contamination=top_n / len(ground_truth))


def top_n_ranges_threshold(ground_truth: np.array, scores: np.array, top_n: Optional[int] = None) -> np.array:
    """
    This function computes a threshold such that there are top_n ranges of anomalies. For this a threshold in the
    given scores is computed. If multiple such thresholds exist, then the smallest threshold is chosen. If no
    threshold exists such that there are top_n ranges, then the (smallest) threshold that gives the most possible
    ranges is chosen. If no **top_n** is given, then the ground truth number of anomalies is used to compute
    the number of ranges.

    :param ground_truth:
    :param scores:
    :param top_n:
    :return:
    """
    if top_n is None:
        top_n = count_nb_ranges(ground_truth)

    if top_n <= 0:
        return np.zeros_like(scores, dtype=np.int16)
    elif top_n >= len(scores):
        return np.ones_like(scores, dtype=np.int16)
    else:
        thresholds = np.sort(np.unique(scores))
        nb_ranges = np.array([count_nb_ranges(scores >= threshold) for threshold in thresholds])
        index_threshold = np.argmax(np.where(nb_ranges <= top_n, nb_ranges, 0))
        return np.array(scores >= thresholds[index_threshold], dtype=np.int16)


def count_nb_ranges(labels) -> int:
    return np.sum(np.diff(labels, prepend=0) == 1)




import numpy as np
import importlib
from typing import Optional, Dict, Union
from pyod.models.base import BaseDetector

from dtaianomaly.anomaly_detection.TimeSeriesAnomalyDetector import TimeSeriesAnomalyDetector
from dtaianomaly.anomaly_detection.utility.TrainType import TrainType
from dtaianomaly.anomaly_detection.utility.Windowing import Windowing

_SUPPORTED_PYOD_ANOMALY_DETECTORS = {
    # key is the name to use when loading, value is the name of the module in PYOD
    'IForest': 'iforest',
    'LOF': 'lof',
    'KNN': 'knn',
    'OCSVM': 'ocsvm',
    'ABOD': 'abod',
    'ECOD': 'ecod'
}


class PyODAnomalyDetector(TimeSeriesAnomalyDetector):

    def __init__(self, pyod_anomaly_detector: Union[BaseDetector, str], windowing: Windowing):
        super().__init__()

        if type(pyod_anomaly_detector) == str:
            # Check if the given anomaly detector is supported
            if pyod_anomaly_detector not in _SUPPORTED_PYOD_ANOMALY_DETECTORS:
                raise ValueError(f"The given anomaly detector '{pyod_anomaly_detector}' is not supported yet, or is not a valid PyODAnomalyDetector!\n"
                                 f"Supported PYODAnomalyDetectors are: {_SUPPORTED_PYOD_ANOMALY_DETECTORS.keys()}")

            # Load the module and class
            module = importlib.import_module(name='pyod.models.' + _SUPPORTED_PYOD_ANOMALY_DETECTORS[pyod_anomaly_detector])
            self.__pyod_anomaly_detector = getattr(module, pyod_anomaly_detector)()
        else:
            self.__pyod_anomaly_detector: BaseDetector = pyod_anomaly_detector
        self.__windowing: Windowing = windowing

    def train_type(self) -> TrainType:
        return TrainType.UNSUPERVISED  # All PyOD anomaly detectors are unsupervised (https://pyod.readthedocs.io/en/latest/api_cc.html#pyod.models.base.BaseDetector.fit)

    def _fit_anomaly_detector(self, trend_data: np.ndarray, labels: Optional[np.array] = None) -> 'PyODAnomalyDetector':
        self.__pyod_anomaly_detector.fit(self.__windowing.create_windows(trend_data))
        return self

    def _compute_decision_scores(self, trend_data: np.ndarray) -> np.array:
        windowed_decision_scores = self.__pyod_anomaly_detector.decision_function(self.__windowing.create_windows(trend_data))
        return self.__windowing.reverse_windowing(windowed_decision_scores)

    @staticmethod
    def load(parameters: Dict[str, any]) -> 'TimeSeriesAnomalyDetector':

        # Check if the given anomaly detector is supported
        if parameters['pyod_model'] not in _SUPPORTED_PYOD_ANOMALY_DETECTORS:
            raise ValueError(f"The given anomaly detector '{parameters['pyod_model']}' is not supported yet, or is not a valid PYODAnomalyDetector!\n"
                             f"Supported PyODAnomalyDetectors are: {_SUPPORTED_PYOD_ANOMALY_DETECTORS.keys()}")

        # Load the module and class
        module = importlib.import_module(name='pyod.models.' + _SUPPORTED_PYOD_ANOMALY_DETECTORS[parameters['pyod_model']])
        pyod_anomaly_detector = getattr(module, parameters['pyod_model'])

        # Initialize the anomaly detector
        return PyODAnomalyDetector(
            pyod_anomaly_detector=pyod_anomaly_detector(**(parameters['pyod_model_parameters'] if 'pyod_model_parameters' in parameters else {})),
            windowing=Windowing(**(parameters['windowing'] if 'windowing' in parameters else {}))
        )




import enum


class TrainType(enum.Enum):
    SUPERVISED = 1
    SEMI_SUPERVISED = 2
    UNSUPERVISED = 3

    def can_solve_train_type_data(self, train_type_data: str) -> bool:

        # Check the input
        if train_type_data not in ['supervised', 'semi-supervised', 'unsupervised']:
            raise ValueError(f"Unknown train type '{train_type_data}'!"
                             f"Valid train types are 'supervised', 'semi-supervised' and 'unsupervised'!")

        # If the train type of this is unsupervised, then it can solve all datasets.
        if self == TrainType.UNSUPERVISED:
            return True

        # If the train type of this is semi-supervised, then it can solve both supervised and semi-supervised datasets.
        if self == TrainType.SEMI_SUPERVISED:
            return train_type_data == 'semi-supervised' or train_type_data == 'supervised'

        # If the train type of this is supervised, then it can only solve supervised datasets.
        if self == TrainType.SUPERVISED:
            return train_type_data == 'supervised'

    def __str__(self):
        return self.name.lower().replace('_', '-')




import pandas as pd
from typing import Union, Optional

from dtaianomaly.data_management import DataManager
from dtaianomaly.anomaly_detection.utility.TrainType import TrainType

from dtaianomaly.workflows.handle_data_configuration import DataConfigurationType, handle_data_configuration
from dtaianomaly.workflows.handle_algorithm_configuration import AlgorithmConfigurationType, handle_algorithm_configuration
from dtaianomaly.workflows.handle_metric import MetricConfigurationType, handle_metric_configuration, metric_configuration_to_names

ConfigurationType = Union[DataConfigurationType, AlgorithmConfigurationType, MetricConfigurationType]


def main(data_manager: DataManager,
         data_configuration: DataConfigurationType,
         algorithm_configuration: AlgorithmConfigurationType,
         metric_configuration: MetricConfigurationType,
         results_path: Optional[str] = None) -> pd.DataFrame:

    data_manager = handle_data_configuration(data_manager, data_configuration)
    algorithm = handle_algorithm_configuration(algorithm_configuration)
    algorithm_train_type = algorithm.train_type()
    results = pd.DataFrame(
        index=pd.MultiIndex.from_tuples(data_manager.get(), names=['collection_name', 'dataset_name']),
        columns=metric_configuration_to_names(metric_configuration)
    )

    for dataset_index in data_manager.get():

        meta_data = data_manager.get_metadata(dataset_index)
        if not algorithm_train_type.can_solve_train_type_data(meta_data['train_type']):
            raise Exception(f"Algorithm type '{algorithm_train_type}' can not solve dataset type '{meta_data['train_type']}'!")

        # For supervised algorithms, the ground truth of the train data is required
        if algorithm_train_type == TrainType.SUPERVISED:
            data_train, ground_truth_train = data_manager.load_raw_data(dataset_index, train=True)
            algorithm.fit(data_train, ground_truth_train)

        # For semi-supervised algorithms, train data is required but not its ground truth
        elif algorithm_train_type == TrainType.SEMI_SUPERVISED:
            data_train, _ = data_manager.load_raw_data(dataset_index, train=True)
            algorithm.fit(data_train)

        # For unsupervised algorithms, use the train data to fit, if available, and otherwise use the test data
        else:
            if meta_data['train_type'] == 'unsupervised':
                data_train, _ = data_manager.load_raw_data(dataset_index, train=False)
            else:
                data_train, _ = data_manager.load_raw_data(dataset_index, train=True)
            algorithm.fit(data_train)

        # Execute on the test data
        data_test, ground_truth_test = data_manager.load_raw_data(dataset_index, train=False)
        predicted_proba = algorithm.predict_proba(data_test)

        # Write away the results
        results.loc[dataset_index] = handle_metric_configuration(metric_configuration, predicted_proba, ground_truth_test)

    # Save the results, if requested
    if results_path is not None:
        results.to_csv(results_path)

    return results




import json
from typing import Dict, Any, Union
from dtaianomaly.data_management import DataManager

DataConfigurationType = Union[Dict[str, Any], str]


def handle_data_configuration(data_manager: DataManager, data_configuration: DataConfigurationType) -> DataManager:

    # Read the data configuration file if it is a string
    if type(data_configuration) is str:
        configuration_file = open(data_configuration, 'r')
        data_configuration = json.load(configuration_file)
        configuration_file.close()

    # Select the datasets that match the given properties
    for select_options in data_configuration['select']:
        data_manager.select(select_options)

    # Return the data manager
    return data_manager




import json
import pandas as pd
import numpy as np
from typing import Dict, Any, Union, List

from dtaianomaly.evaluation.auc_metrics import roc_auc, pr_auc
from dtaianomaly.evaluation.classification_metrics import precision, recall, f1
from dtaianomaly.evaluation.thresholding import fixed_value_threshold, contamination_threshold, top_n_threshold, top_n_ranges_threshold

MetricConfigurationType = Union[Dict[str, Dict[str, Any]], str]


# Functions with parameters: ground truth anomalies, predicted anomaly scores, metric parameters, thresholding function
__SUPPORTED_METRICS = {
    # Area under the curve metrics (do not require thresholding)
    'roc_auc': lambda ground_truth_anomalies, predicted_anomaly_scores, _, __:
        roc_auc(ground_truth_anomalies, predicted_anomaly_scores),
    'pr_auc': lambda ground_truth_anomalies, predicted_anomaly_scores, _, __:
        pr_auc(ground_truth_anomalies, predicted_anomaly_scores),

    # Basic metrics
    'precision': lambda ground_truth_anomalies, predicted_anomaly_scores, _, thresholding:
        precision(ground_truth_anomalies, thresholding(ground_truth_anomalies, predicted_anomaly_scores)),
    'recall': lambda ground_truth_anomalies, predicted_anomaly_scores, _, thresholding:
        recall(ground_truth_anomalies, thresholding(ground_truth_anomalies, predicted_anomaly_scores)),
    'f1': lambda ground_truth_anomalies, predicted_anomaly_scores, _, thresholding:
        f1(ground_truth_anomalies, thresholding(ground_truth_anomalies, predicted_anomaly_scores))
}
__METRICS_WITHOUT_THRESHOLDING = ['roc_auc', 'pr_auc']

# Return a function that takes as input the ground truth anomalies and the predicted anomaly scores, and returns the labels of the predicted anomalies
__SUPPORTED_THRESHOLDING = {
    'fixed_value': lambda params: (lambda ground_truth_anomalies, predicted_anomaly_scores: fixed_value_threshold(ground_truth_anomalies, predicted_anomaly_scores, **params)),
    'contamination': lambda params: (lambda ground_truth_anomalies, predicted_anomaly_scores: contamination_threshold(ground_truth_anomalies, predicted_anomaly_scores, **params)),
    'top_n': lambda params: (lambda ground_truth_anomalies, predicted_anomaly_scores: top_n_threshold(ground_truth_anomalies, predicted_anomaly_scores, **params)),
    'top_n_ranges': lambda params: (lambda ground_truth_anomalies, predicted_anomaly_scores: top_n_ranges_threshold(ground_truth_anomalies, predicted_anomaly_scores, **params))
}


def handle_metric_configuration(metric_configuration: MetricConfigurationType, predicted_proba: np.array, ground_truth: np.array) -> pd.Series:

    # Read the metric configuration file if it is a string
    if type(metric_configuration) is str:
        configuration_file = open(metric_configuration, 'r')
        metric_configuration = json.load(configuration_file)
        configuration_file.close()

    # Compute the metrics
    results = pd.Series(index=metric_configuration_to_names(metric_configuration))
    for metric, metric_parameters in metric_configuration.items():

        # Check if the given metric is supported
        if metric in __SUPPORTED_METRICS:
            metric_name = metric
        else:
            if 'metric_name' not in metric_parameters:
                raise ValueError(f"The metric parameters do not contain a 'metric_name' property for entry with key '{metric}'!\n"
                                 f"Given metric parameters: {metric_parameters}")
            if metric_parameters['metric_name'] not in __SUPPORTED_METRICS:
                raise ValueError(f"The given metric '{metric_parameters['metric_name']}' is not supported yet, or is not a valid metric!\n"
                                 f"Supported metrics are: {__SUPPORTED_METRICS.keys()}")
            metric_name = metric_parameters['metric_name']

        # Check if the given metric requires thresholding (and raise an error if no thresholding information is given if required)
        if metric_name in __METRICS_WITHOUT_THRESHOLDING:
            thresholding_function = None
        else:
            if 'thresholding_strategy' not in metric_parameters:
                raise ValueError(f"The metric parameters do not contain a 'thresholding_strategy' property for entry with key '{metric}'!\n"
                                 f"Given metric parameters: {metric_parameters}")
            if metric_configuration[metric]['thresholding_strategy'] not in __SUPPORTED_THRESHOLDING:
                raise ValueError(f"The given thresholding '{metric_parameters['thresholding_strategy']}' is not supported yet, or is not a valid thresholding!\n"
                                 f"Supported thresholding methods are: {__SUPPORTED_THRESHOLDING.keys()}")

            thresholding_parameters = metric_parameters['thresholding_parameters'] if 'thresholding_parameters' in metric_parameters else {}
            thresholding_function = __SUPPORTED_THRESHOLDING[metric_parameters['thresholding_strategy']](thresholding_parameters)

        # Compute the specific metric
        metric_parameters = metric_configuration[metric]['metric_parameters'] if 'metric_parameters' in metric_configuration[metric] else {}
        results[metric] = __SUPPORTED_METRICS[metric_name](ground_truth, predicted_proba, metric_parameters, thresholding_function)

    return results


def metric_configuration_to_names(metric_configuration: MetricConfigurationType) -> List[str]:
    if type(metric_configuration) is str:
        configuration_file = open(metric_configuration, 'r')
        metric_configuration = json.load(configuration_file)
        configuration_file.close()

    metric_names = []
    for metric, metric_parameters in metric_configuration.items():
        if metric in __SUPPORTED_METRICS:
            metric_names.append(metric)
        else:
            metric_names.append(metric_parameters['metric_name'])
    return list(metric_configuration.keys())




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional, Union, Dict


def plot_data(
        trend_data: pd.DataFrame,
        file_path: Optional[str] = None,
        show_ground_truth: Optional[str] = None) -> plt.Figure:

    # Check if valid value is provided for 'show_ground_truth'
    if show_ground_truth not in [None, 'overlay', 'background']:
        raise ValueError("Parameter 'show_ground_truth' must be one of None, 'overlay', or 'background'!")

    # Identify the attributes and set up the axis
    attributes = trend_data.columns.drop('is_anomaly')
    fig, axs = plt.subplots(nrows=len(attributes), sharex='all')
    axs = [axs] if len(attributes) == 1 else axs

    # Plot each attribute
    for i in range(len(attributes)):
        attribute = attributes[i]
        if show_ground_truth is None:
            _plot_data(axs[i], trend_data[attribute])
        elif show_ground_truth == 'overlay':
            _plot_data_with_anomaly_overlay(axs[i], trend_data[attribute], trend_data['is_anomaly'].values)
        elif show_ground_truth == 'background':
            _add_anomalies_as_background(axs[i], trend_data['is_anomaly'])
            _plot_data(axs[i], trend_data[attribute])

    # Save the figure if requested
    if file_path is not None:
        fig.savefig(file_path)

    return fig


def plot_anomaly_scores(
        trend_data: pd.DataFrame,
        anomaly_scores: Union[np.array, Dict[str, np.array]],
        file_path: Optional[str] = None,
        show_anomaly_scores: str = 'separate',
        show_ground_truth: Optional[str] = None) -> plt.Figure:

    # Check if valid value is provided for 'show_anomaly_scores'
    if show_anomaly_scores not in ['separate', 'overlay']:
        raise ValueError("Parameter 'show_anomaly_scores' must be one of 'separate' or 'overlay'!")

    # Check if valid value is provided for 'show_ground_truth'
    if show_ground_truth not in [None, 'compare', 'overlay', 'background']:
        raise ValueError("Parameter 'show_ground_truth' must be one of None, 'compare', 'overlay', or 'background'!")

    # Check compatibilities of showing anomaly scores and ground truth
    if show_ground_truth == 'compare' and show_anomaly_scores == 'overlay':
        raise ValueError("'show_ground_truth' with value 'compare' and 'show_anomaly_scores' with value 'overlay' are incompatible!"
                         "Cannot show anomaly scores as overlay when comparing to ground truth!")
    if show_ground_truth == 'overlay' and show_anomaly_scores == 'overlay':
        raise ValueError("'show_ground_truth' with value 'overlay' and 'show_anomaly_scores' with value 'overlay' are incompatible!"
                         "Cannot show both ground truth and anomaly score as a overlay!")
    if type(anomaly_scores) == dict and show_anomaly_scores == 'overlay':
        raise ValueError("Cannot show anomaly scores as overlay when multiple anomaly scores are provided!")

    # Format the anomaly scores to a dict with as key the label to show in the plot
    formatted_anomaly_scores = anomaly_scores
    if type(anomaly_scores) != dict:
        formatted_anomaly_scores = {'Anomaly scores': anomaly_scores}

    # Set up the figure
    nb_rows = trend_data.shape[1] - 1
    nb_rows += len(formatted_anomaly_scores) if show_anomaly_scores == 'separate' else 0
    fig, axs = plt.subplots(nrows=nb_rows, sharex='all')
    axs = [axs] if nb_rows == 1 else axs

    # Plot each attribute
    attributes = trend_data.columns.drop('is_anomaly')
    for i in range(len(attributes)):
        attribute = attributes[i]
        if show_anomaly_scores == 'overlay':
            if show_ground_truth == 'background':
                _add_anomalies_as_background(axs[i], trend_data['is_anomaly'])
            _plot_data_with_anomaly_overlay(axs[i], trend_data[attribute], list(formatted_anomaly_scores.values())[0])
        elif show_ground_truth is None:
            _plot_data(axs[i], trend_data[attribute])
        elif show_ground_truth == 'overlay':
            _plot_data_with_anomaly_overlay(axs[i], trend_data[attribute], trend_data['is_anomaly'].values)
        elif show_ground_truth == 'background':
            _add_anomalies_as_background(axs[i], trend_data['is_anomaly'])
            _plot_data(axs[i], trend_data[attribute])

    # Plot the anomaly separate, if requested
    if show_anomaly_scores == 'separate':
        axs_counter = len(attributes)
        for label, specific_anomaly_scores in formatted_anomaly_scores.items():
            axs[axs_counter].set_title(label)
            if show_ground_truth == 'compare':
                axs[axs_counter].plot(specific_anomaly_scores, color='red', label='Predicted')
                axs[axs_counter].plot(trend_data['is_anomaly'], color='green', label='Ground truth')
                axs[axs_counter].legend()
            else:
                axs[axs_counter].plot(specific_anomaly_scores)
            # Increment the counter
            axs_counter += 1

    # Save the figure if requested
    if file_path is not None:
        fig.savefig(file_path)

    return fig


def _plot_data(ax: plt.Axes, trend_data_attribute: pd.Series) -> None:
    trend_data_attribute.plot(ax=ax, title=trend_data_attribute.name)


def _add_anomalies_as_background(ax: plt.Axes, anomaly_labels: pd.Series) -> None:
    anomaly_ranges = np.where(np.diff(anomaly_labels, prepend=0) != 0)[0].reshape(-1, 2)
    for start, end in anomaly_ranges:
        # Mark the point before and after the anomaly as well
        start_inclusive = start - 1
        end_inclusive = end + 1
        ax.axvspan(start_inclusive, end_inclusive, color='red', label='Anomaly', alpha=0.2)


def _plot_data_with_anomaly_overlay(ax: plt.Axes, trend_data_attribute: pd.Series, anomaly_scores: np.array, nb_bins=100) -> None:
    normalized_scores = (anomaly_scores - np.min(anomaly_scores)) / (np.max(anomaly_scores) - np.min(anomaly_scores))
    binned_anomaly_scores = [np.floor(score * nb_bins) / nb_bins for score in normalized_scores]
    colormap = plt.get_cmap('RdYlGn', nb_bins).reversed()
    for i in range(0, trend_data_attribute.shape[0]):
        color = colormap(binned_anomaly_scores[i])
        ax.plot(trend_data_attribute.loc[i:i + 2], c=color)




from .auc_metrics import roc_auc, pr_auc
from .classification_metrics import precision, recall, f1

from .thresholding import fixed_value_threshold, contamination_threshold, top_n_threshold, top_n_ranges_threshold



import numpy as np


class Windowing:

    def __init__(self, window_size: int, reduction: str = 'mean'):
        self.__window_size: int = window_size
        if reduction == 'mean':
            self.__reduction_function = np.mean
        elif reduction == 'median':
            self.__reduction_function = np.median
        elif reduction == 'max':
            self.__reduction_function = np.max
        elif reduction == 'sum':
            self.__reduction_function = np.sum
        else:
            raise ValueError(f"Reduction strategy '{reduction}' is invalid for reverse windowing! Allowed strategies are 'mean', 'median', 'max', and 'sum'.")

    @property
    def window_size(self) -> int:
        return self.__window_size

    def create_windows(self, trend_data: np.ndarray) -> np.ndarray:
        # Add a new dimension (of size 1) to the trend data if only one dimension exists
        if len(trend_data.shape) == 1:
            trend_data = trend_data.reshape(-1, 1)

        nb_windows = trend_data.shape[0] - self.__window_size + 1
        windowed_trend_data = np.empty((nb_windows, self.__window_size * trend_data.shape[1]), dtype=trend_data.dtype)
        for w in range(nb_windows):
            windowed_trend_data[w, :] = trend_data[w:w + self.__window_size, :].flatten()
        return windowed_trend_data

    def reverse_windowing(self, scores: np.array) -> np.array:
        total_time = scores.shape[0] + self.__window_size - 1
        scores_time = np.empty(total_time)

        # The first time points are covered by less than |self.__window_size| windows
        for t in range(self.__window_size):
            scores_time[t] = self.__reduction_function(scores[:t + 1])
        # The middle time points are covered by exactly |self.__window_size| windows
        for t in range(self.__window_size, total_time - self.__window_size + 1):
            scores_time[t] = self.__reduction_function(scores[t - self.__window_size + 1:t + 1])
        # And, the last time points are also covered by less than |self.__window_size| windows
        for t in range(total_time - self.__window_size + 1, total_time):
            scores_time[t] = self.__reduction_function(scores[t - self.__window_size + 1:])

        return scores_time




import numpy as np
from sklearn.metrics import precision_score, recall_score, f1_score


def precision(ground_truth_anomalies: np.array, predicted_anomalies: np.array) -> float:
    return precision_score(ground_truth_anomalies, predicted_anomalies)


def recall(ground_truth_anomalies: np.array, predicted_anomalies: np.array) -> float:
    return recall_score(ground_truth_anomalies, predicted_anomalies)


def f1(ground_truth_anomalies: np.array, predicted_anomalies: np.array) -> float:
    return f1_score(ground_truth_anomalies, predicted_anomalies)



import abc
import numpy as np
import copy
from typing import Optional, Dict
from scipy.stats import binom
from scipy.special import erf

from dtaianomaly.anomaly_detection.utility.TrainType import TrainType


class TimeSeriesAnomalyDetector(abc.ABC):

    def __init__(self):
        self.__decision_scores = None

    @abc.abstractmethod
    def train_type(self) -> TrainType:
        raise NotImplementedError("Abstract method 'is_supervised()' should be implemented by the specific anomaly detector!")

    def fit(self, trend_data: np.ndarray, labels: Optional[np.array] = None) -> 'TimeSeriesAnomalyDetector':
        self.__decision_scores = None
        self._fit_anomaly_detector(trend_data, labels)
        return self

    @abc.abstractmethod
    def _fit_anomaly_detector(self, trend_data: np.ndarray, labels: Optional[np.array] = None):
        raise NotImplementedError("Abstract method 'fit(np.ndarray, Optional[np.array])' should be implemented by the specific anomaly detector!")

    def decision_function(self, trend_data: np.ndarray) -> np.array:
        if self.__decision_scores is None:
            self.__decision_scores = self._compute_decision_scores(trend_data)
        return self.__decision_scores

    @abc.abstractmethod
    def _compute_decision_scores(self, trend_data: np.ndarray) -> np.array:
        raise NotImplementedError("Abstract method 'decision_scores(np.ndarray)' should be implemented by the specific anomaly detector!")

    @staticmethod
    @abc.abstractmethod
    def load(parameters: Dict[str, any]) -> 'TimeSeriesAnomalyDetector':
        raise NotImplementedError("Abstract method 'load()' should be implemented by the specific anomaly detector!")

    def predict_proba(self, trend_data: np.ndarray, normalization: str = 'unify') -> np.array:
        decision_scores = self.decision_function(trend_data)
        return self._normalize(decision_scores, normalization)

    @staticmethod
    def _normalize(decision_scores: np.array, normalization: str) -> np.array:
        # Use min-max normalization to normalize the decision scores
        if normalization == 'min_max':
            min_decision_score = np.min(decision_scores)
            max_decision_score = np.max(decision_scores)
            probability = (decision_scores - min_decision_score) / (max_decision_score - min_decision_score)

        # Use a unifying strategy to normalize the decision scores (https://epubs.siam.org/doi/abs/10.1137/1.9781611972818.2)
        elif normalization == 'unify':
            mean = np.mean(decision_scores)
            std = np.std(decision_scores)
            pre_erf_scores = (decision_scores - mean) / (std * np.sqrt(2))
            erf_scores = erf(pre_erf_scores)
            probability = erf_scores.clip(0, 1).ravel()

        # Raise an exception if an invalid decision score was given
        else:
            raise ValueError(f"Invalid normalization strategy: '{normalization}'!"
                             f"Valid options are: 'unify', 'min_max'")

        # Return the probabilities, aka normalized decision scores,
        return probability

    def predict_confidence(self, trend_data: np.ndarray, contamination: float) -> np.array:
        # paper: https://link.springer.com/chapter/10.1007/978-3-030-67664-3_14
        # github: https://github.com/Lorenzo-Perini/Confidence_AD/tree/master
        n_samples = len(trend_data)
        n_anomalies = int(n_samples * contamination)
        posterior_prob = self.predict_proba(trend_data)
        prediction = posterior_prob > np.quantile(posterior_prob, 1 - contamination)

        conf_func = np.vectorize(lambda p: 1 - binom.cdf(n_samples - n_anomalies, n_samples, p))
        ex_wise_conf = conf_func(posterior_prob)
        np.place(ex_wise_conf, prediction == 0, 1 - ex_wise_conf[prediction == 0])  # if the example is classified as normal, use 1 - confidence.

        return ex_wise_conf

    def deepcopy(self) -> 'TimeSeriesAnomalyDetector':
        return copy.deepcopy(self)




import json
import importlib
from typing import Dict, Any, Union
from dtaianomaly.anomaly_detection import *

AlgorithmConfigurationType = Union[Dict[str, Any], str]


def handle_algorithm_configuration(algorithm_configuration: AlgorithmConfigurationType) -> TimeSeriesAnomalyDetector:

    # Read the algorithm configuration file if it is a string
    if type(algorithm_configuration) is str:
        configuration_file = open(algorithm_configuration, 'r')
        algorithm_configuration = json.load(configuration_file)
        configuration_file.close()

    # Load the specific anomaly detector class
    anomaly_detector_class_object: TimeSeriesAnomalyDetector = getattr(importlib.import_module('dtaianomaly.anomaly_detection'), algorithm_configuration['anomaly_detector'])

    # Load and return the specific anomaly detector instance
    return anomaly_detector_class_object.load(parameters=algorithm_configuration)




from pathlib import Path

import toml
from setuptools import setup, find_packages

pyproject = toml.load('pyproject.toml')

this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()
long_description_content_type = 'text/markdown'

setup(
    name=pyproject["project"]["name"],
    version=pyproject["project"]["version"],
    author=pyproject["project"]["authors"][0]["name"],
    author_email=pyproject["project"]["authors"][0]["email"],
    description=pyproject["project"]["description"],
    keywords=pyproject["project"]["keywords"],
    license=pyproject["project"]["license"],
    project_urls=pyproject["project"]["urls"],
    url=pyproject["project"]["urls"]["repository"],

    long_description=long_description,
    long_description_content_type=long_description_content_type,

    python_requires=pyproject["project"]["requires-python"],
    install_requires=pyproject["project"]["dependencies"],
    include_package_data=True,
    packages=find_packages(include=['dtaianomaly', 'dtaianomaly.*']),

    setup_requires=pyproject["build-system"]["requires"],
    zip_safe=False,
)



import os
import pandas as pd
import numpy as np
from typing import Tuple, Optional, Dict, List, Union

DatasetIndex = Tuple[str, str]


_COLUMN_NAMES = {
    'train_path', 'test_path', 'dataset_type', 'datetime_index', 'split_at',
    'train_type', 'train_is_normal', 'input_type', 'length', 'dimensions',
    'contamination', 'num_anomalies', 'min_anomaly_length', 'median_anomaly_length',
    'max_anomaly_length', 'mean', 'stddev', 'trend', 'stationarity', 'period_size'
}


# Inspired by Timeeval (https://github.com/HPI-Information-Systems/TimeEval/tree/main)
class DataManager:

    def __init__(self, data_dir: str, datasets_index_file: str = 'datasets.csv'):

        # Check if the data directory exists
        if not os.path.exists(data_dir):
            raise ValueError(f"The given data directory '{data_dir}' does not exist!")
        self.__data_dir: str = data_dir

        # Check if the index file exists
        if not os.path.exists(data_dir + '/' + datasets_index_file):
            raise ValueError(f"There is no file '{datasets_index_file}' in the directory '{data_dir}'!")
        self.__datasets_index_file: str = datasets_index_file

        # Check if the index file contains the correct index columns
        with open(self.__data_dir + '/' + self.__datasets_index_file, 'r') as file:
            columns = file.readline().strip().split(',')
            if 'collection_name' not in columns and 'dataset_name' not in columns:
                raise IndexError(f"The dataset index file does not contain columns 'collection_name' and 'dataset_name'!")

            # Check if the columns are correct
            columns.remove('collection_name')
            columns.remove('dataset_name')
            if set(columns) != _COLUMN_NAMES:
                raise ValueError(f"The dataset index file does not contain the correct columns!\n"
                                 f"Columns in the file: {columns}\n"
                                 f"Required columns: {_COLUMN_NAMES}")

        # Read the dataset index file
        self.__datasets_index: pd.DataFrame = pd.read_csv(self.__data_dir + '/' + self.__datasets_index_file, index_col=['collection_name', 'dataset_name'])

        # Read the dataset index file
        self.__selected_datasets: pd.Series = pd.Series(index=self.__datasets_index.index, data=False)

    def clear(self) -> None:
        # Set all values in the selected datasets to False
        self.__selected_datasets[:] = False

    def select(self, dataset_properties: Optional[Dict[str, any]] = None) -> None:

        # If no dataset properties are given, select all datasets
        if dataset_properties is None:
            self.__selected_datasets = pd.Series(index=self.__datasets_index.index, data=True)
            return

        # Keep track of the datasets that match all the given properties
        newly_selected_datasets = np.ones(self.__datasets_index.shape[0], dtype=bool)

        # Filter the datasets that do not match the given properties
        for dataset_property, value in dataset_properties.items():
            # Make sure the given property is valid
            if not (dataset_property in self.__datasets_index.columns or dataset_property in self.__datasets_index.index.names):
                raise ValueError(f"The dataset property '{dataset_property}' is not a valid property for a dataset!\n"
                                 f"Valid properties are: {self.__datasets_index.columns.tolist()}")

            # Filter the index explicitly
            if dataset_property == 'collection_name':
                newly_selected_datasets &= self.__datasets_index.index.get_level_values('collection_name').isin([value] if isinstance(value, str) else value)
            elif dataset_property == 'dataset_name':
                newly_selected_datasets &= self.__datasets_index.index.get_level_values('dataset_name').isin([value] if isinstance(value, str) else value)

            # Filter the remaining values implicitly, based on the type of the column
            else:
                # For boolean properties, there should only be given one value: either True or False
                if self.__datasets_index[dataset_property].dtype == bool:
                    if not isinstance(value, bool):
                        raise ValueError(f"The dataset property '{dataset_property}' is a boolean, but {value} is not a boolean!")
                    newly_selected_datasets &= (self.__datasets_index.loc[:, dataset_property] == value)

                # For number-like properties, either a single number (for exact match) or a list of two numbers (for a range) should be given
                elif self.__datasets_index[dataset_property].dtype == np.int64 or self.__datasets_index[dataset_property].dtype == np.float64:
                    if isinstance(value, list) or isinstance(value, tuple):
                        if len(value) != 2:
                            raise ValueError(f"For dataset property '{dataset_property}', the value should either express a number (exact match), or a list of two numbers (min and max)!")
                        if not (isinstance(value[0], int) or isinstance(value[0], float)) or not (isinstance(value[1], int) or isinstance(value[1], float)):
                            raise ValueError(f"Both attributes in a ranged value for dataset property '{dataset_property}' should be either an int or a float!")
                        newly_selected_datasets &= (self.__datasets_index.loc[:, dataset_property] >= value[0]) & (self.__datasets_index.loc[:, dataset_property] <= value[1])
                    else:
                        if not (isinstance(value, int) or isinstance(value, float)):
                            raise ValueError(f"The dataset property '{dataset_property}' is a number (float or int), but {value} is not a number!")
                        newly_selected_datasets &= (self.__datasets_index.loc[:, dataset_property] == value)

                # Otherwise, either a single value (for exact match) or a list of values (for multiple exact matches) should be given
                else:
                    iterable_value = isinstance(value, list) or isinstance(value, tuple) or isinstance(value, set)
                    newly_selected_datasets &= self.__datasets_index.loc[:, dataset_property].isin(value if iterable_value else [value])

        # Set the flags of the selected datasets to True
        self.__selected_datasets |= newly_selected_datasets

    def filter_available_datasets(self) -> None:
        for dataset_index in self.__selected_datasets.index:

            # Skip datasets that are already not selected
            if not self.__selected_datasets[dataset_index]:
                continue

            # Obtain the metadata, in particular to get the train and test path
            metadata = self.get_metadata(dataset_index)

            # For train data, also check if there should be training data (or if there is only test data)
            if pd.notna(self.get_metadata(dataset_index)['train_path']):
                if not os.path.exists(self.__data_dir + '/' + metadata['train_path']):
                    self.__selected_datasets[dataset_index] = False
                    continue  # Don't have to check the test data anymore

            # The test data must exist
            if not os.path.exists(self.__data_dir + '/' + metadata['test_path']):
                self.__selected_datasets[dataset_index] = False

    def get(self, index: Optional[int] = None) -> Union[List[DatasetIndex], DatasetIndex]:
        selected_datasets = self.__selected_datasets[self.__selected_datasets].index
        if index is None:
            return selected_datasets.tolist()
        else:
            if index < 0 or index >= selected_datasets.shape[0]:
                raise IndexError(f"Invalid index of a dataset given to the DataManager: '{index}'!"
                                 f"Only {selected_datasets.shape[0]} datasets are selected, thus '0 <= index < {selected_datasets.shape[0]}' must be satisfied")
            return selected_datasets[index]

    def get_metadata(self, dataset_index: DatasetIndex) -> pd.Series:
        self.check_index_exists(dataset_index)
        return self.__datasets_index.loc[dataset_index, :]

    def load(self, dataset_index: DatasetIndex, train: bool = False) -> pd.DataFrame:
        self.check_index_selected(dataset_index)
        path = self.check_data_exists(dataset_index, train)
        return pd.read_csv(path, index_col='timestamp')

    def load_raw_data(self, dataset_index: DatasetIndex, train: bool = False) -> Tuple[np.ndarray, np.ndarray]:
        data_df = self.load(dataset_index, train)
        return data_df[data_df.columns].drop(columns='is_anomaly').values, data_df['is_anomaly'].values

    def check_index_exists(self, dataset_index: DatasetIndex) -> None:
        if dataset_index not in self.__datasets_index.index:
            raise ValueError(f"The dataset '{dataset_index}' does not exist!")

    def check_index_selected(self, dataset_index: DatasetIndex) -> None:
        self.check_index_exists(dataset_index)
        if not self.__selected_datasets.loc[dataset_index]:
            raise ValueError(f"The dataset '{dataset_index}' has not been selected!")

    def check_data_exists(self, dataset_index: DatasetIndex, train: bool = False) -> str:
        if train and pd.isna(self.__datasets_index.loc[dataset_index, 'train_path']):
            raise ValueError(f"Train data is requested, but there is no train data for the dataset with index '{dataset_index}'!")

        path = (self.__data_dir + '/' + self.__datasets_index.loc[dataset_index, 'train_path' if train else 'test_path'])
        if not os.path.exists(path):
            raise FileNotFoundError(f"The {'train' if train else 'test'} data of dataset '{dataset_index}' does not exist!")
        return path

    def add_dataset(self,
                    collection_name: str,
                    dataset_name: str,
                    test_data: pd.DataFrame,
                    test_path: str,
                    dataset_type: str,
                    train_type: str,
                    train_is_normal: bool,
                    trend: str,
                    stationarity: str,
                    train_data: Optional[pd.DataFrame] = None,
                    train_path: Optional[str] = None,
                    period_size: Optional[int] = None,
                    split_at: Optional[int] = None) -> None:

        # Check if the dataset already exists
        dataset_index = (collection_name, dataset_name)
        if dataset_index in self.__datasets_index.index:
            raise ValueError(f"There already exists a dataset with index '{dataset_index}'!")

        # Check the test data and train data if given
        check_data(test_data, test_path, 'test')
        if train_data is not None and train_path is not None:
            check_data(train_data, train_path, 'train')
            if set(train_data) != set(test_data):  # Includes the 'is_anomaly' column
                raise ValueError(f"The train data and test data should have the same number of attributes!")
        elif train_data is not None and train_path is None:
            raise ValueError(f"Train data is given, but there is no train data path given!")
        elif train_data is None and train_path is not None:
            raise ValueError(f"Train data path is given, but there is no train data given!")

        length_anomaly_sequences = []
        current_length = 0
        for anomaly_label in test_data['is_anomaly']:
            if anomaly_label == 1:
                # If the label is 1, then the current anomaly sequence increased in length
                current_length += 1
            else:
                # If the label is 0, then a sequence has ended if the current length is larger than 0
                if current_length > 0:
                    length_anomaly_sequences.append(current_length)
                    current_length = 0
        if current_length > 0:
            length_anomaly_sequences.append(current_length)

        # Create a new row for the dataset index
        new_row = pd.Series(index=_COLUMN_NAMES, data={
            'train_path': train_path,
            'test_path': test_path,
            'dataset_type': dataset_type,
            'datetime_index': set(test_data.index) != set(range(test_data.shape[0])),
            'split_at': split_at,
            'train_type': train_type,
            'train_is_normal': train_is_normal,
            'input_type': 'multivariate' if test_data.shape[1] > 2 else 'univariate',  # '2' to take 'is_anomaly' into account
            'length': test_data.shape[0],
            'dimensions': test_data.shape[1] - 1,  # Exclude the 'is_anomaly' column
            'contamination': test_data['is_anomaly'].sum() / test_data.shape[0],
            'num_anomalies': len(length_anomaly_sequences),
            'min_anomaly_length': min(length_anomaly_sequences),
            'median_anomaly_length': np.median(length_anomaly_sequences),
            'max_anomaly_length': max(length_anomaly_sequences),
            'mean': np.mean(test_data.mean(axis=0).drop('is_anomaly')),
            'stddev': np.mean(test_data.std(axis=0).drop('is_anomaly')),
            'trend': trend,
            'stationarity': stationarity,
            'period_size': period_size
        })

        # Add the row to the dataset index
        self.__datasets_index.loc[dataset_index, :] = new_row
        self.__selected_datasets.at[dataset_index] = False

        # Save the index and the data
        self.__datasets_index.to_csv(self.__data_dir + '/' + self.__datasets_index_file)
        if not os.path.exists(self.__data_dir + '/' + os.path.dirname(test_path)):
            os.makedirs(self.__data_dir + '/' + os.path.dirname(test_path))
        test_data.to_csv(self.__data_dir + '/' + test_path)
        if train_data is not None:
            if not os.path.exists(os.path.dirname(self.__data_dir + '/' + train_path)):
                os.makedirs(os.path.dirname(self.__data_dir + '/' + train_path))
            train_data.to_csv(self.__data_dir + '/' + train_path)

    def remove_dataset(self, dataset_index: DatasetIndex) -> None:
        self.check_index_exists(dataset_index)
        metadata = self.get_metadata(dataset_index)

        # Update the index file
        self.__datasets_index.drop(index=dataset_index, inplace=True)
        self.__datasets_index.to_csv(self.__data_dir + '/' + self.__datasets_index_file)

        # Update the selected datasets
        self.__selected_datasets.drop(index=dataset_index, inplace=True)

        # Remove the data files
        if pd.notna(metadata['train_path']):
            os.remove(self.__data_dir + '/' + metadata['train_path'])
        os.remove(self.__data_dir + '/' + metadata['test_path'])


def check_data(data: pd.DataFrame, path: str, name: str) -> None:
    # Check if the path already exists
    if os.path.exists(path):
        raise ValueError(f"The given {name} path '{path}' already exist!")
    # Check if the index has the correct name
    if data.index.name != 'timestamp':
        raise ValueError(f"The index of the given {name} data does not have the name 'timestamp', but the name '{data.index.name}'!")
    # Check if there is an is_anomaly column
    if 'is_anomaly' not in data.columns:
        raise ValueError(f"The given {name} data does not contain the column 'is_anomaly'!")
    # Check if the 'is_anomaly' column only contains binary labels (both integers and floats)
    if not data['is_anomaly'].isin([0, 1, 0.0, 1.0, True, False]).all():
        raise ValueError(f"The 'is_anomaly' column of the {name} data should only contain binary labels ('0' or '1'), but contains values!")




import numpy as np
import sklearn


def roc_auc(ground_truth_anomalies: np.array, predicted_anomaly_scores: np.array) -> float:
    return sklearn.metrics.roc_auc_score(ground_truth_anomalies, predicted_anomaly_scores)


def pr_auc(ground_truth_anomalies: np.array, predicted_anomaly_scores: np.array) -> float:
    precision, recall, _ = sklearn.metrics.precision_recall_curve(ground_truth_anomalies, predicted_anomaly_scores)
    return sklearn.metrics.auc(recall, precision)
